theory Auxiliary
  imports Probability "~~/src/HOL/Library/Binomial"
begin

lemma (in pair_sigma_finite) Fubini':
  assumes *: "(\<lambda>(x, y). f x y) \<in> borel_measurable (M1 \<Otimes>\<^sub>M M2)"
  shows "(\<integral>\<^sup>+y. \<integral>\<^sup>+x. f x y \<partial>M1 \<partial>M2) = (\<integral>\<^sup>+x. \<integral>\<^sup>+y. f x y \<partial>M2 \<partial>M1)"
  using Fubini[OF *] by simp

lemma emeasure_countable_UNION:
  assumes sets: "\<And>i. i \<in> I \<Longrightarrow> X i \<in> sets M" and I: "countable I" 
  assumes disj: "disjoint_family_on X I"
  shows "emeasure M (UNION I X) = (\<integral>\<^sup>+i. emeasure M (X i) \<partial>count_space I)"
proof cases
  assume "finite I" with sets disj show ?thesis
    apply (subst setsum_emeasure[symmetric])
    apply (auto intro!: setsum_cong simp add: max_def subset_eq positive_integral_count_space_finite emeasure_nonneg)
    done
next
  assume f: "\<not> finite I"
  then have [intro]: "I \<noteq> {}" by auto
  from f have "bij_betw (from_nat_into I) UNIV I"
    using bij_betw_from_nat_into[OF I] by simp
  then have "(\<Union>i\<in>I. X i) = (\<Union>i. (X \<circ> from_nat_into I) i)"
    unfolding SUP_def image_compose by (simp add: bij_betw_def)
  then have "emeasure M (UNION I X) = emeasure M (\<Union>i. X (from_nat_into I i))"
    by simp
  also have "\<dots> = (\<Sum>i. emeasure M (X (from_nat_into I i)))"
    using disj
    apply (intro suminf_emeasure[symmetric])
    apply (auto simp: disjoint_family_on_def intro!: sets from_nat_into[OF `I \<noteq> {}`])
    apply (erule_tac x="from_nat_into I m" in ballE)
    apply (auto simp: disjoint_family_on_def intro!: sets from_nat_into[OF `I \<noteq> {}`])
    apply (erule_tac x="from_nat_into I n" in ballE)
    apply (auto simp: disjoint_family_on_def from_nat_into_inj_infinite[OF I f]
        intro!: sets from_nat_into[OF `I \<noteq> {}`])
    done
  also have "\<dots> = (\<Sum>n. \<integral>\<^sup>+i. emeasure M (X i) * indicator {from_nat_into I n} i \<partial>count_space I)"
  proof (intro arg_cong[where f=suminf] ext)
    fix i
    have eq: "{a \<in> I. 0 < emeasure M (X a) * indicator {from_nat_into I i} a}
     = (if 0 < emeasure M (X (from_nat_into I i)) then {from_nat_into I i} else {})"
     apply (auto simp: ereal_zero_less_0_iff indicator_def from_nat_into[OF `I \<noteq> {}`])
     apply (simp add: one_ereal_def zero_ereal_def)
     done
    have "(\<integral>\<^sup>+ ia. emeasure M (X ia) * indicator {from_nat_into I i} ia \<partial>count_space I) =
      (if 0 < emeasure M (X (from_nat_into I i)) then emeasure M (X (from_nat_into I i)) else 0)"
      by (subst positive_integral_count_space) (simp_all add: eq)
    also have "\<dots> = emeasure M (X (from_nat_into I i))"
      by (simp add: less_le emeasure_nonneg)
    finally show "emeasure M (X (from_nat_into I i)) =
         \<integral>\<^sup>+ ia. emeasure M (X ia) * indicator {from_nat_into I i} ia \<partial>count_space I" ..
  qed
  also have "\<dots> = (\<integral>\<^sup>+i. emeasure M (X i) \<partial>count_space I)"
    apply (subst positive_integral_suminf[symmetric])
    apply (auto simp: ereal_0_le_mult emeasure_nonneg intro!: positive_integral_cong)
  proof -
    fix x assume "x \<in> I"
    then have "(\<Sum>i. emeasure M (X x) * indicator {from_nat_into I i} x) = (\<Sum>i\<in>{to_nat_on I x}. emeasure M (X x) * indicator {from_nat_into I i} x)"
      by (intro suminf_finite) (auto simp: indicator_def I f)
    also have "\<dots> = emeasure M (X x)"
      by (simp add: I f `x\<in>I`)
    finally show "(\<Sum>i. emeasure M (X x) * indicator {from_nat_into I i} x) = emeasure M (X x)" .
  qed
  finally show ?thesis .
qed

lemma (in finite_measure) integrable_measure: 
  assumes I: "disjoint_family_on X I" "countable I"
  shows "integrable (count_space I) (\<lambda>i. measure M (X i))"
proof -
  have "(\<integral>\<^sup>+i. measure M (X i) \<partial>count_space I) = (\<integral>\<^sup>+i. measure M (if X i \<in> sets M then X i else {}) \<partial>count_space I)"
    by (auto intro!: positive_integral_cong measure_notin_sets)
  also have "\<dots> = measure M (\<Union>i\<in>I. if X i \<in> sets M then X i else {})"
    using I unfolding emeasure_eq_measure[symmetric]
    by (subst emeasure_countable_UNION) (auto simp: disjoint_family_on_def)
  finally show ?thesis
    by (auto intro!: integrable_nonneg simp: measure_nonneg)
qed

lemma sigma_finite_measure_count_space:
  assumes A: "countable A"
  shows "sigma_finite_measure (count_space A)"
proof
  show "\<exists>F::nat \<Rightarrow> 'a set. range F \<subseteq> sets (count_space A) \<and> (\<Union>i. F i) = space (count_space A) \<and>
     (\<forall>i. emeasure (count_space A) (F i) \<noteq> \<infinity>)"
     using A
     apply (intro exI[of _ "\<lambda>i. {from_nat_into A i} \<inter> A"])
     apply auto
     apply (rule_tac x="to_nat_on A x" in exI)
     apply simp
     done
qed

definition smallest :: "(nat \<Rightarrow> 'a) \<Rightarrow> 'a set \<Rightarrow> nat \<Rightarrow> bool" where
  "smallest \<omega> S n \<longleftrightarrow> (\<forall>i<n. \<omega> i \<notin> S) \<and> \<omega> n \<in> S"

lemma smallest_unique: "smallest \<omega> S n \<Longrightarrow> smallest \<omega> S m \<Longrightarrow> m = n"
  by (rule ccontr) (auto simp: smallest_def simp: neq_iff)

lemma smallestD1: "smallest \<omega> S n \<Longrightarrow> \<omega> n \<in> S"
  by (auto simp: smallest_def)

lemma smallestD2: "smallest \<omega> S n \<Longrightarrow> i < n \<Longrightarrow> \<omega> i \<notin> S"
  by (auto simp: smallest_def)

lemma smallestI: "\<omega> n \<in> S \<Longrightarrow> smallest \<omega> S (LEAST i. \<omega> i \<in> S)"
  unfolding smallest_def by (rule LeastI2_wellorder[where a=n]) auto

lemma LeastI_smallest: "smallest \<omega> S n \<Longrightarrow> (LEAST i. \<omega> i \<in> S) = n"
  by (blast intro: smallest_unique smallestI dest: smallestD1)

lemma (in complete_lattice) Inf_eq: "(\<And>a. a \<in> A \<Longrightarrow> \<exists>b\<in>B. b \<le> a) \<Longrightarrow> (\<And>b. b \<in> B \<Longrightarrow> \<exists>a\<in>A. a \<le> b) \<Longrightarrow>  Inf A = Inf B"
  by (intro antisym Inf_greatest) (blast intro: Inf_lower2 dest: assms)+

lemma one_not_le_zero_ereal[simp]: "\<not> (1 \<le> (0::ereal))"
  by (simp add: one_ereal_def zero_ereal_def)

lemma eventually_at_left_1: "(\<And>z::real. 0 < z \<Longrightarrow> z < 1 \<Longrightarrow> P z) \<Longrightarrow> eventually P (at_left 1)"
  unfolding eventually_at_left by (auto intro!: exI[of _ 0])

lemma (in prob_space) integral_le_const:
  "integrable M f \<Longrightarrow> (AE x in M. f x \<le> c) \<Longrightarrow> (\<integral>x. f x \<partial>M) \<le> c"
  using integral_mono_AE[of M f "\<lambda>x. c"] prob_space by simp

lemma (in prob_space) positive_integral_ge_const:
  "(AE x in M. c \<le> f x) \<Longrightarrow> c \<le> (\<integral>\<^sup>+x. f x \<partial>M)"
  using positive_integral_mono_AE[of "\<lambda>x. c" f M] emeasure_space_1
  by (simp add: positive_integral_const_If split: split_if_asm)

lemma (in -) sums_suminf_ereal: "f sums x \<Longrightarrow> (\<Sum>i. ereal (f i)) = ereal x"
  by (metis sums_ereal sums_unique)

lemma (in -) suminf_ereal': "summable f \<Longrightarrow> (\<Sum>i. ereal (f i)) = ereal (\<Sum>i. f i)"
  by (metis sums_ereal sums_unique summable_def)

lemma sets_pair_countable:
  assumes "countable S1" "countable S2"
  assumes M: "sets M = Pow S1" and N: "sets N = Pow S2"
  shows "sets (M \<Otimes>\<^sub>M N) = Pow (S1 \<times> S2)"
proof auto
  fix x a b assume x: "x \<in> sets (M \<Otimes>\<^sub>M N)" "(a, b) \<in> x"
  from sets.sets_into_space[OF x(1)] x(2)
    sets_eq_imp_space_eq[of N "count_space S2"] sets_eq_imp_space_eq[of M "count_space S1"] M N
  show "a \<in> S1" "b \<in> S2"
    by (auto simp: space_pair_measure)
next
  fix X assume X: "X \<subseteq> S1 \<times> S2"
  then have "countable X"
    by (metis countable_subset `countable S1` `countable S2` countable_SIGMA)
  have "X = (\<Union>(a, b)\<in>X. {a} \<times> {b})" by auto
  also have "\<dots> \<in> sets (M \<Otimes>\<^sub>M N)"
    using X
    by (safe intro!: sets.countable_UN' `countable X` subsetI pair_measureI) (auto simp: M N)
  finally show "X \<in> sets (M \<Otimes>\<^sub>M N)" .
qed

lemma positive_intergal_count_space_nat:
  fixes f :: "nat \<Rightarrow> ereal"
  assumes nonneg: "\<And>i. 0 \<le> f i"
  shows "(\<integral>\<^sup>+i. f i \<partial>count_space UNIV) = (\<Sum>i. f i)"
proof -
  have "(\<integral>\<^sup>+i. f i \<partial>count_space UNIV) =
    (\<integral>\<^sup>+i. (\<Sum>j. f j * indicator {j} i) \<partial>count_space UNIV)"
  proof (intro positive_integral_cong)
    fix i
    have "f i = (\<Sum>j\<in>{i}. f j * indicator {j} i)"
      by simp
    also have "\<dots> = (\<Sum>j. f j * indicator {j} i)"
      by (rule suminf_finite[symmetric]) auto
    finally show "f i = (\<Sum>j. f j * indicator {j} i)" .
  qed
  also have "\<dots> = (\<Sum>j. (\<integral>\<^sup>+i. f j * indicator {j} i \<partial>count_space UNIV))"
    by (rule positive_integral_suminf) (auto simp: nonneg)
  also have "\<dots> = (\<Sum>j. f j)"
    by (simp add: nonneg positive_integral_cmult_indicator one_ereal_def[symmetric])
  finally show ?thesis .
qed

lemma emeasure_countable_singleton:
  assumes sets: "\<And>x. x \<in> X \<Longrightarrow> {x} \<in> sets M" and X: "countable X"
  shows "emeasure M X = (\<integral>\<^sup>+x. emeasure M {x} \<partial>count_space X)"
proof -
  have X_eq: "X = (\<Union>a. if a \<in> to_nat_on X ` X then {from_nat_into X a} else {})"
    using X
    apply (auto split: split_if_asm cong: rev_conj_cong)
    apply (rule_tac x="to_nat_on X x" in exI)
    apply simp
    done

  have eq: "\<And>n. {x. from_nat_into X n = x \<and> x \<in> X \<and> n \<in> to_nat_on X ` X} =
    (if n \<in> to_nat_on X ` X then {from_nat_into X n} else {})"
    using X by auto

  have "(\<integral>\<^sup>+x. emeasure M {x} \<partial>count_space X) = 
    (\<integral>\<^sup>+x. (\<Sum>n. emeasure M {from_nat_into X n} * indicator {x\<in>X. n \<in> to_nat_on X ` X \<and> from_nat_into X n = x} x) \<partial>count_space X)"
  proof (intro positive_integral_cong)
    fix x assume "x \<in> space (count_space X)"
    then have "emeasure M {x} = (\<Sum>n\<in>{to_nat_on X x}. emeasure M {from_nat_into X n} * indicator {x\<in>X. n \<in> to_nat_on X ` X \<and> from_nat_into X n = x} x)"
      using X by simp
    also have "\<dots> = (\<Sum>n. emeasure M {from_nat_into X n} * indicator {x\<in>X. n \<in> to_nat_on X ` X \<and> from_nat_into X n = x} x)"
      using X by (intro suminf_finite[symmetric]) (auto simp: indicator_def)
    finally show "emeasure M {x} =
                    (\<Sum>n. emeasure M {from_nat_into X n} * indicator {x\<in>X. n \<in> to_nat_on X ` X \<and> from_nat_into X n = x} x)" .
  qed
  also have "\<dots> = (\<Sum>n. (\<integral>\<^sup>+x. emeasure M {from_nat_into X n} * indicator {x\<in>X. n \<in> to_nat_on X ` X \<and> from_nat_into X n = x} x \<partial>count_space X))"
    by (simp_all add: emeasure_nonneg positive_integral_suminf)
  also have "\<dots> = (\<Sum>n. emeasure M (if n \<in> to_nat_on X ` X then {from_nat_into X n} else {}))"
    using X
    by (subst positive_integral_cmult_indicator) (auto simp add: emeasure_nonneg eq one_ereal_def[symmetric] intro!: arg_cong[where f=suminf] ext)
  also have "\<dots> = emeasure M X"
    using `countable X`
    apply (subst (4) X_eq)
    apply (subst suminf_emeasure[symmetric]) 
    apply (auto simp: disjoint_family_on_def intro!: arg_cong[where f=suminf] sets)
    done
  finally show ?thesis ..
qed


lemma measure_eqI_countable:
  assumes [simp]: "sets M = Pow A" "sets N = Pow A" and A: "countable A"
  assumes eq: "\<And>a. a \<in> A \<Longrightarrow> emeasure M {a} = emeasure N {a}"
  shows "M = N"
proof (rule measure_eqI)
  fix X assume "X \<in> sets M"
  then have X: "X \<subseteq> A" by auto
  moreover with A have "countable X" by (auto dest: countable_subset)
  ultimately have
    "emeasure M X = (\<integral>\<^sup>+a. emeasure M {a} \<partial>count_space X)"
    "emeasure N X = (\<integral>\<^sup>+a. emeasure N {a} \<partial>count_space X)"
    by (auto intro!: emeasure_countable_singleton)
  moreover have "(\<integral>\<^sup>+a. emeasure M {a} \<partial>count_space X) = (\<integral>\<^sup>+a. emeasure N {a} \<partial>count_space X)"
    using X by (intro positive_integral_cong eq) auto
  ultimately show "emeasure M X = emeasure N X"
    by simp
qed simp

lemma comb_seq_id: "comb_seq n \<omega> \<omega>' n = \<omega>' 0"
  by (simp add: comb_seq_def)

lemma comb_seq_gt: "i < j \<Longrightarrow> comb_seq i \<omega> \<omega>' j = \<omega>' (j - i)"
  unfolding comb_seq_def by simp

lemma comb_seq_Suc2: "i \<le> n \<Longrightarrow> comb_seq (Suc i) \<omega> \<omega>' n = nat_case (\<omega> i) \<omega>' (n - i)"
  unfolding comb_seq_def by (auto simp: diff_Suc split: nat.split)

lemma disjCI2: "(\<not> Q \<Longrightarrow> P) \<Longrightarrow> Q \<or> P"
  by blast

lemma eventually_nhds_top:
  fixes P :: "'a :: {order_top, linorder_topology} \<Rightarrow> bool"
  assumes "(b::'a) < top"
  shows "eventually P (nhds top) \<longleftrightarrow> (\<exists>b<top. (\<forall>z. b < z \<longrightarrow> P z))"
  unfolding eventually_nhds
proof safe
  fix S :: "'a set" assume "open S" "top \<in> S"
  note open_left[OF this `b < top`]
  moreover assume "\<forall>s\<in>S. P s"
  ultimately show "\<exists>b<top. \<forall>z>b. P z"
    by (auto simp: subset_eq Ball_def)
next
  fix b assume "b < top" "\<forall>z>b. P z"
  then show "\<exists>S. open S \<and> top \<in> S \<and> (\<forall>xa\<in>S. P xa)"
    by (intro exI[of _ "{b <..}"]) auto
qed

lemma suminf_ereal_finite: "summable f \<Longrightarrow> (\<Sum>i. ereal (f i)) \<noteq> \<infinity>"
  by (auto simp: sums_ereal[symmetric] summable_def sums_unique[symmetric])

lemma tendsto_PInfty_eq_at_top:
  "((\<lambda>z. ereal (f z)) ---> \<infinity>) F \<longleftrightarrow> (LIM z F. f z :> at_top)"
  unfolding tendsto_PInfty filterlim_at_top_dense by simp

lemma mult_left_le:
  "(c::_::linordered_semidom) \<le> 1 \<Longrightarrow> 0 \<le> a \<Longrightarrow> a * c \<le> a"
  using mult_left_mono[of c 1 a] by simp

lemma mult_le_one: 
  "a \<le> (1::_::linordered_semidom) \<Longrightarrow> 0 \<le> b \<Longrightarrow> b \<le> 1 \<Longrightarrow> a * b \<le> 1"
  using mult_mono[of a 1 b 1] by simp

lemma suminf_ereal_offset_le:
  fixes f :: "nat \<Rightarrow> ereal"
  assumes f: "\<And>i. 0 \<le> f i"
  shows "(\<Sum>i. f (i + k)) \<le> suminf f"
proof -
  have "(\<lambda>n. \<Sum>i<n. f (i + k)) ----> (\<Sum>i. f (i + k))"
    using summable_sums[OF summable_ereal_pos] by (simp add: sums_def atLeast0LessThan f)
  moreover have "(\<lambda>n. \<Sum>i<n. f i) ----> (\<Sum>i. f i)"
    using summable_sums[OF summable_ereal_pos] by (simp add: sums_def atLeast0LessThan f)
  then have "(\<lambda>n. \<Sum>i<n + k. f i) ----> (\<Sum>i. f i)"
    by (rule LIMSEQ_ignore_initial_segment)
  ultimately show ?thesis
  proof (rule LIMSEQ_le, safe intro!: exI[of _ k])
    fix n assume "k \<le> n"
    have "(\<Sum>i<n. f (i + k)) = (\<Sum>i<n. (f \<circ> (\<lambda>i. i + k)) i)"
      by simp
    also have "\<dots> = (\<Sum>i\<in>(\<lambda>i. i + k) ` {..<n}. f i)"
      by (subst setsum_reindex) auto
    also have "\<dots> \<le> setsum f {..<n + k}"
      by (intro setsum_mono3) (auto simp: f)
    finally show "(\<Sum>i<n. f (i + k)) \<le> setsum f {..<n + k}" .
  qed
qed

lemma Lim_within_LIMSEQ:
  fixes a :: "'a::first_countable_topology"
  assumes "\<forall>S. (\<forall>n. S n \<noteq> a \<and> S n \<in> T) \<and> S ----> a \<longrightarrow> (\<lambda>n. X (S n)) ----> L"
  shows "(X ---> L) (at a within T)"
  using assms unfolding tendsto_def [where l=L]
  by (simp add: sequentially_imp_eventually_within)

lemma Ex_incseq_LIMSEQ:
  fixes b a :: "_ :: {linorder_topology, first_countable_topology, dense_linorder}"
  assumes "b < a"
  shows "\<exists>F. (\<forall>i. F i < a) \<and> incseq F \<and> F ----> a"
proof -
  from Topological_Spaces.countable_basis[of a] obtain A where 
    A: "\<And>i. open (A i)" "\<And>i. a \<in> A i" "\<And>F. \<forall>n. F n \<in> A n \<Longrightarrow> F ----> a"
    by metis
  obtain L where L: "\<And>i. L i < a" "\<And>i. {L i<..a} \<subseteq> A i"
    using open_left[OF A(1, 2) `b < a`] by metis
  then have L_A: "\<And>i x. L i < x \<Longrightarrow> x \<le> a \<Longrightarrow> x \<in> A i" by (auto simp: subset_eq)

  let ?P = "\<lambda>n b x. L n < x \<and> b \<le> x \<and> x < a"
  def r \<equiv> "nat_rec (SOME x. ?P 0 b x) (\<lambda>n b. SOME x. ?P (Suc n) b x)"
  { fix b n assume "b < a"
    with L(1)[of n] have "max b (L n) < a" by auto
    from dense[OF this] obtain z where z: "max b (L n) < z" "z < a" by auto
    then have "\<exists>z. ?P n b z"
      by auto }
  note * = this

  { fix n have "L n < r n \<and> nat_case b r n \<le> r n \<and> r n < a"
    proof (induct n)
      case 0 show ?case
        by (simp add: r_def) (rule someI_ex[OF *, OF `b < a`])
    next
      case (Suc n) show ?case
        unfolding r_def nat_rec_Suc
        unfolding r_def[symmetric]
        apply (simp add: )
        apply (rule_tac someI_ex[OF *])
        using Suc by simp
    qed }
  note * = this
  with L(2) have "\<And>i. r i < a" "\<And>i. r i \<in> A i"
    by (auto intro!: L_A less_imp_le)
  moreover have "incseq r"
    proof (rule incseq_SucI) fix n from *[of "Suc n"] show "r n \<le> r (Suc n)" by simp qed
  ultimately show ?thesis
    using A(3) by auto
qed

lemma sequentially_imp_eventually_at_left:
  fixes a :: "_ :: {no_bot, dense_linorder, linorder_topology, first_countable_topology}"
  assumes [simp]: "b < a"
  assumes *: "\<And>f. (\<And>n. b < f n) \<Longrightarrow> (\<And>n. f n < a) \<Longrightarrow> incseq f \<Longrightarrow> f ----> a \<Longrightarrow> eventually (\<lambda>n. P (f n)) sequentially"
  shows "eventually P (at_left a)"
proof (rule ccontr)
  assume "\<not> eventually P (at_left a)"
  then obtain F where F: "\<And>b. b < a \<Longrightarrow> b < F b" "\<And>b. b < a \<Longrightarrow> F b < a" "\<And>b. b < a \<Longrightarrow> \<not> P (F b)"
    unfolding eventually_at_left[of P] by metis

  obtain G where G: "\<And>i. G i < a" "incseq G" "G ----> a"
    using Ex_incseq_LIMSEQ[OF `b < a`] by auto

  have less_F: "\<And>x y. x \<le> y \<Longrightarrow> y < a \<Longrightarrow> x < F y"
    by (metis le_less_trans F(1))

  def r \<equiv> "nat_rec (F (max b (G 0))) (\<lambda>n b. F (max b (G (Suc n))))"
  then have r: "r 0 = F (max b (G 0))" "\<And>n. r (Suc n) = F (max (r n) (G (Suc n)))"
    unfolding r_def nat_rec_0 nat_rec_Suc unfolding r_def[symmetric] by auto

  { fix n have "r n < a" by (induct n) (simp_all add: r F G) }
  note r_less = this

  { fix n have "G n \<le> r n"
      by (induct n) (simp_all add: r less_F G less_imp_le r_less) }
  note G_r = this

  { fix n have "b < r n"
      by (induct n) (simp_all add: r_less r G less_F min_max.le_supI1) }
  note b = this

  { fix n have "\<not> P (r n)" by (induct n) (simp_all add: r r_less F G) }
  note not_P = this

  have "incseq r"
    by (rule incseq_SucI) (simp add: less_F G r_less r less_imp_le)

  have "r ----> a"
  proof (intro order_tendstoI)
    fix x assume "a < x"
    from r_less this have "\<And>n. r n < x" by (rule less_trans)
    then show "eventually (\<lambda>n. r n < x) sequentially" by simp
  next
    fix x assume "x < a"
    note order_tendstoD(1)[OF G(3) this]
    moreover have "eventually (\<lambda>n. G n \<le> r n) sequentially"
      using G_r by simp
    ultimately show "eventually (\<lambda>n. x < r n) sequentially"
      by eventually_elim simp
  qed
  hence "eventually (\<lambda>n. P (r n)) sequentially"
    using b r_less `incseq r` by (intro *) auto
  with not_P show False by simp
qed

lemma tendsto_at_left_sequentially:
  fixes a :: "_ :: {no_bot, dense_linorder, linorder_topology, first_countable_topology}"
  assumes "b < a"
  assumes *: "\<And>S. (\<And>n. S n < a) \<Longrightarrow> (\<And>n. b < S n) \<Longrightarrow> incseq S \<Longrightarrow> S ----> a \<Longrightarrow> (\<lambda>n. X (S n)) ----> L"
  shows "(X ---> L) (at_left a)"
  using assms unfolding tendsto_def [where l=L]
  by (simp add: sequentially_imp_eventually_at_left)

lemma positive_integral_LIMSEQ:
  assumes f: "incseq f" "\<And>i. f i \<in> borel_measurable M" "\<And>n x. 0 \<le> f n x"
    and u: "\<And>x. (\<lambda>i. f i x) ----> u x"
  shows "(\<lambda>n. integral\<^sup>P M (f n)) ----> integral\<^sup>P M u"
proof -
  have "(\<lambda>n. integral\<^sup>P M (f n)) ----> (SUP n. integral\<^sup>P M (f n))"
    using f by (intro LIMSEQ_SUP[of "\<lambda>n. integral\<^sup>P M (f n)"] incseq_positive_integral)
  also have "(SUP n. integral\<^sup>P M (f n)) = integral\<^sup>P M (\<lambda>x. SUP n. f n x)"
    using f by (intro positive_integral_monotone_convergence_SUP[symmetric])
  also have "integral\<^sup>P M (\<lambda>x. SUP n. f n x) = integral\<^sup>P M (\<lambda>x. u x)"
    using f by (subst SUP_Lim_ereal[OF _ u]) (auto simp: incseq_def le_fun_def)
  finally show ?thesis .
qed

lemma summable_power_series:
  fixes z :: real
  assumes le_1: "\<And>i. f i \<le> 1" and nonneg: "\<And>i. 0 \<le> f i" and z: "0 \<le> z" "z < 1"
  shows "summable (\<lambda>i. f i * z^i)"
proof (rule summable_comparison_test[OF _ summable_geometric])
  show "norm z < 1" using z by (auto simp: less_imp_le)
  show "\<And>n. \<exists>N. \<forall>na\<ge>N. norm (f na * z ^ na) \<le> z ^ na"
    using z by (auto intro!: exI[of _ 0] mult_left_le_one_le simp: abs_mult nonneg power_abs less_imp_le le_1)
qed

lemma power_series_tendsto_at_left:
  assumes nonneg: "\<And>i. 0 \<le> f i" and summable: "\<And>z. 0 \<le> z \<Longrightarrow> z < 1 \<Longrightarrow> summable (\<lambda>n. f n * z^n)"
  shows "((\<lambda>z. ereal (\<Sum>n. f n * z^n)) ---> (\<Sum>n. ereal (f n))) (at_left (1::real))"
proof (intro tendsto_at_left_sequentially)
  show "0 < (1::real)" by simp
  fix S :: "nat \<Rightarrow> real" assume S: "\<And>n. S n < 1" "\<And>n. 0 < S n" "S ----> 1" "incseq S"
  then have S_nonneg: "\<And>i. 0 \<le> S i" by (auto intro: less_imp_le)

  have "(\<lambda>i. (\<integral>\<^sup>+n. f n * S i^n \<partial>count_space UNIV)) ----> (\<integral>\<^sup>+n. ereal (f n) \<partial>count_space UNIV)"
  proof (rule positive_integral_LIMSEQ)
    show "incseq (\<lambda>i n. ereal (f n * S i^n))"
      using S by (auto intro!: mult_mono power_mono nonneg simp: incseq_def le_fun_def less_imp_le)
    fix n have "(\<lambda>i. ereal (f n * S i^n)) ----> ereal (f n * 1^n)"
      by (intro tendsto_intros lim_ereal[THEN iffD2] S)
    then show "(\<lambda>i. ereal (f n * S i^n)) ----> ereal (f n)"
      by simp
  qed (auto simp: S_nonneg intro!: mult_nonneg_nonneg nonneg)
  also have "(\<lambda>i. (\<integral>\<^sup>+n. f n * S i^n \<partial>count_space UNIV)) = (\<lambda>i. \<Sum>n. f n * S i^n)"
    by (subst positive_intergal_count_space_nat)
       (intro ext suminf_ereal mult_nonneg_nonneg nonneg S_nonneg ereal_less_eq(5)[THEN iffD2]
              zero_le_power suminf_ereal_finite summable S)+
  also have "(\<integral>\<^sup>+n. ereal (f n) \<partial>count_space UNIV) = (\<Sum>n. ereal (f n))"
    by (simp add: positive_intergal_count_space_nat nonneg)
  finally show "(\<lambda>n. ereal (\<Sum>na. f na * S n ^ na)) ----> (\<Sum>n. ereal (f n))" .
qed

lemma measurable_card[measurable]:
  fixes S :: "'a \<Rightarrow> nat set"
  assumes [measurable]: "\<And>i. {x\<in>space M. i \<in> S x} \<in> sets M"
  shows "(\<lambda>x. card (S x)) \<in> measurable M (count_space UNIV)"
  unfolding measurable_count_space_eq2_countable
proof (safe intro!: UNIV_I)
  fix n show "(\<lambda>x. card (S x)) -` {n} \<inter> space M \<in> sets M"
  proof (cases n)
    case 0
    then have "(\<lambda>x. card (S x)) -` {n} \<inter> space M = {x\<in>space M. infinite (S x) \<or> (\<forall>i. i \<notin> S x)}"
      by auto
    also have "\<dots> \<in> sets M"
      by measurable
    finally show ?thesis .
  next
    case (Suc i)
    then have "(\<lambda>x. card (S x)) -` {n} \<inter> space M =
      (\<Union>F\<in>{A\<in>{A. finite A}. card A = n}. {x\<in>space M. (\<forall>i. i \<in> S x \<longleftrightarrow> i \<in> F)})"
      unfolding set_eq_iff[symmetric] Collect_bex_eq[symmetric] by (auto intro: card_ge_0_finite)
    also have "\<dots> \<in> sets M"
      by (intro sets.countable_UN' countable_Collect countable_Collect_finite) auto
    finally show ?thesis .
  qed
qed

lemma ereal_right_mult_cong: 
  fixes a b c d :: ereal
  shows "c = d \<Longrightarrow> (d \<noteq> 0 \<Longrightarrow> a = b) \<Longrightarrow> c * a = d * b"
  by (cases "d = 0") simp_all

theorem binomial_ring:
  fixes a b :: "'a :: {comm_ring_1, power}"
  shows "(a+b)^n = (\<Sum>k=0..n. of_nat (n choose k) * a^k * b^(n-k))"
proof (induct n)
  case 0 thus ?case by simp
next
  case (Suc n)
  have decomp: "{0..n+1} = {0} \<union> {n+1} \<union> {1..n}"
    by (auto simp add:atLeastAtMost_def atLeast_def atMost_def)
  have decomp2: "{0..n} = {0} \<union> {1..n}"
    by (auto simp add:atLeastAtMost_def atLeast_def atMost_def)
  have "(a+b)^(n+1) = (a+b) * (\<Sum>k=0..n. of_nat (n choose k) * a^k * b^(n-k))"
    using Suc by simp
  also have "\<dots> =  a*(\<Sum>k=0..n. of_nat (n choose k) * a^k * b^(n-k)) +
                   b*(\<Sum>k=0..n. of_nat (n choose k) * a^k * b^(n-k))"
    by (simp add: field_simps)
  also have "\<dots> = (\<Sum>k=0..n. of_nat (n choose k) * a^(k+1) * b^(n-k)) +
                  (\<Sum>k=0..n. of_nat (n choose k) * a^k * b^(n-k+1))"
    by (simp add: setsum_right_distrib mult_ac)
  also have "\<dots> = (\<Sum>k=0..n. of_nat (n choose k) * a^k * b^(n+1-k)) +
                  (\<Sum>k=1..n+1. of_nat (n choose (k - 1)) * a^k * b^(n+1-k))"
    by (simp add:setsum_shift_bounds_cl_Suc_ivl Suc_diff_le
             del:setsum_cl_ivl_Suc)
  also have "\<dots> = a^(n+1) + b^(n+1) +
                  (\<Sum>k=1..n. of_nat (n choose (k - 1)) * a^k * b^(n+1-k)) +
                  (\<Sum>k=1..n. of_nat (n choose k) * a^k * b^(n+1-k))"
    by (simp add: decomp2)
  also have "\<dots> = a^(n+1) + b^(n+1) + (\<Sum>k=1..n. of_nat (n+1 choose k) * a^k * b^(n+1-k))"
    by (simp add: field_simps setsum_addf binomial.simps)
  also have "\<dots> = (\<Sum>k=0..n+1. of_nat (n+1 choose k) * a^k * b^(n+1-k))"
    using decomp by simp
  finally show ?case by simp
qed

lemma LIMSEQ_root: "(\<lambda>n. root n n) ----> 1"
proof -
  def x \<equiv> "\<lambda>n. root n n - 1"
  have "x ----> sqrt 0"
  proof (rule tendsto_sandwich[OF _ _ tendsto_const])
    show "(\<lambda>x. sqrt (2 / x)) ----> sqrt 0"
      by (intro tendsto_intros tendsto_divide_0[OF tendsto_const] filterlim_mono[OF filterlim_real_sequentially])
         (simp_all add: at_infinity_eq_at_top_bot)
    { fix n :: nat assume "2 < n"
      have "1 + (real (n - 1) * n) / 2 * x n^2 = 1 + of_nat (n choose 2) * x n^2"
        using `2 < n` unfolding gbinomial_def binomial_gbinomial
        by (simp add: atLeast0AtMost atMost_Suc field_simps real_of_nat_diff numeral_2_eq_2 real_eq_of_nat[symmetric])
      also have "\<dots> \<le> (\<Sum>k\<in>{0, 2}. of_nat (n choose k) * x n^k)"
        by (simp add: x_def)
      also have "\<dots> \<le> (\<Sum>k=0..n. of_nat (n choose k) * x n^k)"
        using `2 < n` by (intro setsum_mono2) (auto intro!: mult_nonneg_nonneg zero_le_power simp: x_def le_diff_eq)
      also have "\<dots> = (x n + 1) ^ n"
        by (simp add: binomial_ring)
      also have "\<dots> = n"
        using `2 < n` by (simp add: x_def)
      finally have "real (n - 1) * (real n / 2 * (x n)\<^sup>2) \<le> real (n - 1) * 1"
        by simp
      then have "(x n)\<^sup>2 \<le> 2 / real n"
        using `2 < n` unfolding mult_le_cancel_left by (simp add: field_simps)
      from real_sqrt_le_mono[OF this] have "x n \<le> sqrt (2 / real n)"
        by simp }
    then show "eventually (\<lambda>n. x n \<le> sqrt (2 / real n)) sequentially"
      by (auto intro!: exI[of _ 3] simp: eventually_sequentially)
    show "eventually (\<lambda>n. sqrt 0 \<le> x n) sequentially"
      by (auto intro!: exI[of _ 1] simp: eventually_sequentially le_diff_eq x_def)
  qed
  from tendsto_add[OF this tendsto_const[of 1]] show ?thesis
    by (simp add: x_def)
qed

lemma LIMSEQ_root_const:
  assumes "0 < c"
  shows "(\<lambda>n. root n c) ----> 1"
proof -
  { fix c :: real assume "1 \<le> c"
    def x \<equiv> "\<lambda>n. root n c - 1"
    have "x ----> 0"
    proof (rule tendsto_sandwich[OF _ _ tendsto_const])
      show "(\<lambda>n. c / n) ----> 0"
        by (intro tendsto_divide_0[OF tendsto_const] filterlim_mono[OF filterlim_real_sequentially])
           (simp_all add: at_infinity_eq_at_top_bot)
      { fix n :: nat assume "1 < n"
        have "1 + x n * n = 1 + of_nat (n choose 1) * x n^1"
          using `1 < n` unfolding gbinomial_def binomial_gbinomial by (simp add: real_eq_of_nat[symmetric])
        also have "\<dots> \<le> (\<Sum>k\<in>{0, 1}. of_nat (n choose k) * x n^k)"
          by (simp add: x_def)
        also have "\<dots> \<le> (\<Sum>k=0..n. of_nat (n choose k) * x n^k)"
          using `1 < n` `1 \<le> c` by (intro setsum_mono2) (auto intro!: mult_nonneg_nonneg zero_le_power simp: x_def le_diff_eq)
        also have "\<dots> = (x n + 1) ^ n"
          by (simp add: binomial_ring)
        also have "\<dots> = c"
          using `1 < n` `1 \<le> c` by (simp add: x_def)
        finally have "x n \<le> c / n"
          using `1 \<le> c` `1 < n` by (simp add: field_simps) }
      then show "eventually (\<lambda>n. x n \<le> c / n) sequentially"
        by (auto intro!: exI[of _ 3] simp: eventually_sequentially)
      show "eventually (\<lambda>n. 0 \<le> x n) sequentially"
        using `1 \<le> c` by (auto intro!: exI[of _ 1] simp: eventually_sequentially le_diff_eq x_def)
    qed
    from tendsto_add[OF this tendsto_const[of 1]] have "(\<lambda>n. root n c) ----> 1"
      by (simp add: x_def) }
  note ge_1 = this

  show ?thesis
  proof cases
    assume "1 \<le> c" with ge_1 show ?thesis by blast
  next
    assume "\<not> 1 \<le> c"
    with `0 < c` have "1 \<le> 1 / c"
      by simp
    then have "(\<lambda>n. 1 / root n (1 / c)) ----> 1 / 1"
      by (intro tendsto_divide tendsto_const ge_1 `1 \<le> 1 / c` one_neq_zero)
    then show ?thesis
      by (rule filterlim_cong[THEN iffD1, rotated 3])
         (auto intro!: exI[of _ 1] simp: eventually_sequentially real_root_divide)
  qed
qed

lemma root_test_convergence:
  fixes f :: "nat \<Rightarrow> 'a::banach"
  assumes f: "(\<lambda>n. root n (norm (f n))) ----> x" -- "could be weakened to lim sup"
  assumes "x < 1"
  shows "summable f"
proof -
  have "0 \<le> x"
    by (rule LIMSEQ_le[OF tendsto_const f]) (auto intro!: exI[of _ 1])
  from `x < 1` obtain z where z: "x < z" "z < 1"
    by (metis dense)
  from f `x < z`
  have "eventually (\<lambda>n. root n (norm (f n)) < z) sequentially"
    by (rule order_tendstoD)
  then have "eventually (\<lambda>n. norm (f n) \<le> z^n) sequentially"
    using eventually_ge_at_top
  proof eventually_elim
    fix n assume less: "root n (norm (f n)) < z" and n: "1 \<le> n"
    from power_strict_mono[OF less, of n] n
    show "norm (f n) \<le> z ^ n"
      by simp
  qed
  then show "summable f"
    unfolding eventually_sequentially
    using z `0 \<le> x` by (auto intro!: summable_comparison_test[OF _  summable_geometric])
qed

lemma (in -) indicator_eq_0_iff: "indicator A x = (0::'a::zero_neq_one) \<longleftrightarrow> x \<notin> A"
  by (auto split: split_indicator)

lemma (in -) lebesgue_integral_point_measure_finite_indicator:
  "finite A \<Longrightarrow> A \<subseteq> S \<Longrightarrow> (\<And>a. a \<in> S \<Longrightarrow> 0 \<le> f a) \<Longrightarrow> integral\<^sup>L (point_measure S f) (\<lambda>x. g x * indicator A x) = (\<Sum>a\<in>A. f a * g a)"
  unfolding point_measure_def
  apply (simp_all add: AE_count_space integral_density point_measure_def)
  apply (subst  lebesgue_integral_count_space_finite_support)
  apply (auto simp add: indicator_eq_0_iff intro!: setsum_mono_zero_cong_left)
  done

lemma atMost_Suc_insert_0: "{.. Suc n} = insert 0 (Suc ` {.. n})"
  apply (auto simp: image_iff Ball_def)
  apply (case_tac x)
  apply auto
  done


lemma (in -) inverse_nonneg_iff: "0 \<le> inverse (x :: ereal) \<longleftrightarrow> 0 \<le> x \<or> x = -\<infinity>"
  by (cases x) auto

lemma (in -) sums_setsum:
  fixes f :: "'i \<Rightarrow> nat \<Rightarrow> 'a::real_normed_field"
  shows "(\<And>i. i \<in> I \<Longrightarrow> (f i) sums (x i)) \<Longrightarrow> (\<lambda>n. \<Sum>i\<in>I. f i n) sums (\<Sum>i\<in>I. x i)"
  by (induct I rule: infinite_finite_induct) (auto intro!: sums_add)

lemma (in -) suminf_setsum:
  fixes f :: "'i \<Rightarrow> nat \<Rightarrow> 'a::real_normed_field"
  shows "(\<And>i. i \<in> I \<Longrightarrow> summable (f i)) \<Longrightarrow> (\<Sum>n. \<Sum>i\<in>I. f i n) = (\<Sum>i\<in>I. \<Sum>n. f i n)"
  using sums_unique[OF sums_setsum, OF summable_sums, of I f] by simp

lemma (in -) summable_setsum:
  fixes f :: "'i \<Rightarrow> nat \<Rightarrow> 'a::real_normed_field"
  shows "(\<And>i. i \<in> I \<Longrightarrow> summable (f i)) \<Longrightarrow> summable (\<lambda>n. \<Sum>i\<in>I. f i n)"
  using sums_summable[OF sums_setsum[OF summable_sums, of I f]] .


lemma smallest_comb_seq: "smallest (comb_seq (Suc i) \<omega> \<omega>') D i \<longleftrightarrow> smallest \<omega> D i"
  by (simp add: smallest_def comb_seq_Suc2 comb_seq_less)

lemma comb_seq_Suc_less: "i < n \<Longrightarrow> comb_seq (Suc i) \<omega> \<omega>' n = \<omega>' (n - Suc i)"
  unfolding comb_seq_def by simp

lemma integral_bound: "f \<in> borel_measurable M \<Longrightarrow> (AE x in M. 0 \<le> f x) \<Longrightarrow> integral\<^sup>P M f \<le> ereal r \<Longrightarrow> integral\<^sup>L M f \<le> r"
  apply (simp add: lebesgue_integral_def)
  apply (subst (2) positive_integral_0_iff_AE[THEN iffD2])
  apply simp_all
  apply (cases "\<integral>\<^sup>+ x. ereal (f x) \<partial>M")
  apply simp_all
  done


lemma PiE_UNIV: "PiE UNIV X = Pi UNIV X"
  by auto

lemma sets_nat_seq:
  fixes S :: "'a set"
  defines "cyl \<omega> n \<equiv> { \<omega>' \<in> UNIV \<rightarrow> S. \<forall>i\<le>n. \<omega>' i = \<omega> i }"
  assumes S: "countable S" and "s \<in> S"
  shows "sets (\<Pi>\<^sub>M i::nat\<in>UNIV. count_space S) = sigma_sets (UNIV \<rightarrow> S) { cyl \<omega> n | \<omega> n. \<forall>i. \<omega> i \<in> S }"
  unfolding sets_PiM PiE_UNIV space_count_space
proof (rule sigma_sets_eqI)
  fix A :: "(nat \<Rightarrow> 'a) set" assume "A \<in> prod_algebra UNIV (\<lambda>i. count_space S)"
  then obtain J E where E: "\<And>j. j \<in> J \<Longrightarrow> E j \<subseteq> S" and J: "finite J" "J \<noteq> {}"
    and A: "A = prod_emb UNIV (\<lambda>s. count_space S) J (PiE J E)"
    by (elim prod_algebraE) auto
  then have [simp]: "\<And>j. j\<in>J \<Longrightarrow> j \<le> Max J" by simp
  def F \<equiv> "\<lambda>i. if i \<in> J then E i else S"
  with E have F: "\<And>j. F j \<subseteq> S" by auto
  from countable_subset[OF this S]
  have [intro]: "\<And>j. countable (F j)" .
  have "A = \<Union> ((\<lambda>\<omega>. cyl \<omega> (Max J)) ` PiE {.. Max J} F )"
    apply (clarsimp simp add: A prod_emb_def Pi_iff PiE_UNIV intro!: set_eqI, safe)
    apply (rule_tac x="restrict x {.. Max J}" in bexI)
    apply (auto simp: cyl_def F_def split: split_if_asm)
    done
  also have "\<dots> \<in> sigma_sets (UNIV \<rightarrow> S) { cyl \<omega> n | \<omega> n. \<forall>i. \<omega> i \<in> S }"
    using F
    apply (intro sigma_sets_UNION countable_image countable_PiE sigma_sets.Basic)
    apply (auto simp: PiE_iff)
    apply (rule_tac x="\<lambda>i. if i \<le> Max J then \<omega> i else s" in exI)
    apply (auto intro!: exI[of _ "Max J"] simp: cyl_def `s\<in>S`)
    done
  finally show "A \<in> sigma_sets (UNIV \<rightarrow> S) { cyl \<omega> n | \<omega> n. \<forall>i. \<omega> i \<in> S }" .
next
  fix A assume "A \<in> {cyl \<omega> n |\<omega> n. \<forall>i. \<omega> i \<in> S}"
  then obtain \<omega> n where "A = {\<omega>'\<in>space (\<Pi>\<^sub>M i\<in>UNIV. count_space S). \<forall>i\<le>n. \<omega>' i = \<omega> i}"
    by (auto simp: space_PiM cyl_def PiE_UNIV)
  also have "\<dots> \<in> sets (\<Pi>\<^sub>M i::nat\<in>UNIV. count_space S)"
    by simp
  also have "\<dots> = sigma_sets (UNIV \<rightarrow> S) (prod_algebra UNIV (\<lambda>i. count_space S))"
    by (simp add: sets_PiM PiE_UNIV)
  finally show "A \<in> sigma_sets (UNIV \<rightarrow> S) (prod_algebra UNIV (\<lambda>i. count_space S))" .
qed

lemma PiM_eq:
  assumes "sets M' = sets (PiM I M)" "sets N' = sets (PiM I M)"
  assumes eq: "\<And>J F. finite J \<Longrightarrow> J \<subseteq> I \<Longrightarrow> (\<And>j. F j \<in> sets (M j)) \<Longrightarrow>
    emeasure M' (prod_emb I M J (\<Pi>\<^sub>E j\<in>J. F j)) = emeasure N' (prod_emb I M J (\<Pi>\<^sub>E j\<in>J. F j))"
  assumes fin: "emeasure N' (space (PiM I M)) \<noteq> \<infinity>"
  shows "M' = N'"
proof (rule measure_eqI_generator_eq[symmetric, OF Int_stable_prod_algebra prod_algebra_sets_into_space])
  have "sets (PiM I M) = sigma_sets (\<Pi>\<^sub>E i\<in>I. space (M i)) (prod_algebra I M)"
    by (rule sets_PiM)
  then show "sets M' = sigma_sets (\<Pi>\<^sub>E i\<in>I. space (M i)) (prod_algebra I M)"
    "sets N' = sigma_sets (\<Pi>\<^sub>E i\<in>I. space (M i)) (prod_algebra I M)"
    unfolding `sets M' = sets (PiM I M)` `sets N' = sets (PiM I M)` by simp_all

  def J \<equiv> "if I = {} then {} else {SOME i. i \<in> I}"
  then have "J \<subseteq> I"
    by (auto intro: someI)
  def A \<equiv> "\<lambda>n::nat. prod_emb I M J (PIE j:J. space (M j))"

  show "range A \<subseteq> prod_algebra I M"
    unfolding A_def J_def image_subset_iff by (intro ballI prod_algebraI) (auto intro: someI)
  have A_eq: "\<And>i. A i = (\<Pi>\<^sub>E i\<in>I. space (M i))"
    unfolding A_def using `J \<subseteq> I` by (force simp: prod_emb_def PiE_iff)
  then show "(\<Union>i. A i) = (\<Pi>\<^sub>E i\<in>I. space (M i))"
    by simp
  
  show "\<And>i. emeasure N' (A i) \<noteq> \<infinity>"
    unfolding A_eq using fin by (simp add: space_PiM)
next
  fix X assume X: "X \<in> prod_algebra I M"
  then obtain J E where X: "X = prod_emb I M J (PIE j:J. E j)"
    and J: "finite J" "J \<subseteq> I" "\<And>j. j \<in> J \<Longrightarrow> E j \<in> sets (M j)"
    by (force elim!: prod_algebraE)
  moreover
  have "emeasure N' (prod_emb I M J (PIE j:J. if j \<in> J then E j else {})) =
    emeasure M' (prod_emb I M J (PIE j:J. if j \<in> J then E j else {}))"
    using J by (intro eq[symmetric]) auto
  ultimately show "emeasure N' X = emeasure M' X"
    by (simp cong: PiE_cong)
qed

lemma nat_seq_eq:
  fixes M :: "nat \<Rightarrow> 'a measure"
  assumes "sets M' = sets (PiM UNIV M)" "sets N' = sets (PiM UNIV M)"
  assumes eq: "\<And>n F. (\<And>i. F i \<in> sets (M i)) \<Longrightarrow>
    emeasure M' {\<omega>\<in>space (PiM UNIV M). \<forall>i\<le>n. \<omega> i \<in> F i} =
    emeasure N' {\<omega>\<in>space (PiM UNIV M). \<forall>i\<le>n. \<omega> i \<in> F i}"
  assumes fin: "emeasure N' (space (PiM UNIV M)) \<noteq> \<infinity>"
  shows "M' = N'"
proof (rule PiM_eq[OF assms(1, 2) _ assms(4)])
  fix J :: "nat set" and F assume "finite J" and [simp]: "\<And>j. F j \<in> sets (M j)"
  def n \<equiv> "if J = {} then 0 else Max J"
  with `finite J` have "\<And>j. j \<in> J \<Longrightarrow> j \<le> n"
    by auto
  then have *: "prod_emb UNIV M J (Pi\<^sub>E J F) =
    {\<omega>\<in>space (PiM UNIV M). \<forall>i\<le>n. \<omega> i \<in> (if i \<in> J then F i else space (M i))}"
    by (auto simp: prod_emb_def PiE_iff space_PiM)
  show "emeasure M' (prod_emb UNIV M J (Pi\<^sub>E J F)) = emeasure N' (prod_emb UNIV M J (Pi\<^sub>E J F))"
    unfolding * by (intro eq) auto
qed

lemma nat_eq_cyl:
  fixes M N :: "(nat \<Rightarrow> 's) measure" and S :: "'s set"
  defines "S_seq \<equiv> (\<Pi>\<^sub>M (i::nat)\<in>UNIV. count_space S)"
  defines "cyl \<equiv> \<lambda>n \<omega>'. {\<omega>\<in>space S_seq. \<forall>i\<le>n. \<omega> i = \<omega>' i}"
  assumes S: "countable S"
  assumes sets_eq: "sets M = sets S_seq" "sets N = sets S_seq"
  assumes eq: "\<And>n \<omega>'. (\<And>i. \<omega>' i \<in> S) \<Longrightarrow> emeasure M {\<omega>\<in>space M. \<forall>i\<le>n. \<omega> i = \<omega>' i} =
    emeasure N {\<omega>\<in>space N. \<forall>i\<le>n. \<omega> i = \<omega>' i}"
  assumes not_inf: "emeasure N (space S_seq) \<noteq> \<infinity>"
  shows "M = N"
proof (rule nat_seq_eq[OF sets_eq[unfolded S_seq_def] _ not_inf[unfolded S_seq_def]], fold S_seq_def)
  fix n :: nat and F :: "nat \<Rightarrow> 's set" assume F: "\<And>i. F i \<in> sets (count_space S)"
  let ?C = "cyl n"

  def X \<equiv> "(\<Pi>\<^sub>E i \<in> {..< Suc n}. F i)"
  then have X_iff[simp]: "\<And>x. x \<in> X \<longleftrightarrow> x \<in> extensional {..< Suc n} \<and> (\<forall>i\<le>n. x i \<in> F i)"
    by (auto simp: PiE_iff)
  
  show "emeasure M {\<omega> \<in> space S_seq. \<forall>i\<le>n. \<omega> i \<in> F i} = emeasure N {\<omega> \<in> space S_seq. \<forall>i\<le>n. \<omega> i \<in> F i}"
  proof cases
    assume not_empty: "\<forall>i\<le>n. F i \<noteq> {}"
    then have [simp]: "X \<noteq> {}"
      unfolding X_def by (auto simp: PiE_eq_empty_iff)

    from not_empty obtain s where "s \<in> F 0"
      by auto
    with F have "s \<in> S" by auto
    
    { fix n :: nat and \<omega>' assume "\<And>i. i \<le> n \<Longrightarrow> \<omega>' i \<in> S"
      moreover have "cyl n \<omega>' = cyl n (\<lambda>i. if i \<le> n then \<omega>' i else s)" (is "_ = cyl _ ?\<omega>")
        by (auto simp: cyl_def)
      moreover note eq[of ?\<omega> n] `s\<in>S`
      ultimately have "emeasure M (cyl n \<omega>') = emeasure N (cyl n \<omega>')"
        by (auto simp: cyl_def sets_eq[THEN sets_eq_imp_space_eq]) }
    note eq = this

    have [simp]: "countable X"
      using F countable_subset[OF _ S] unfolding X_def by (intro countable_PiE) auto
    
    { fix \<omega>1 \<omega>2 x assume *: "\<omega>1 \<in> extensional {..< Suc n}" "\<omega>2 \<in> extensional {..< Suc n}" "x \<in> ?C \<omega>1" "x \<in> ?C \<omega>2"
      { fix i have "\<omega>1 i = \<omega>2 i"
          using * by (cases "i < Suc n") (auto simp: cyl_def extensional_def) }
      then have "\<omega>1 = \<omega>2"
        by auto }
    then have disjoint: "disjoint_family (\<lambda>i. if i \<in> to_nat_on X ` X then ?C (from_nat_into X i) else {})"
      unfolding disjoint_family_on_def by auto

    have *: "{\<omega> \<in> space S_seq. \<forall>i\<le>n. \<omega> i \<in> F i} =
        (\<Union>i. if i \<in> to_nat_on X`X then ?C (from_nat_into X i) else {})"
      apply (auto simp: cyl_def space_PiM PiE_iff split: split_if_asm)
      apply (rule_tac x="to_nat_on X (restrict x {..< Suc n})" in exI)
      apply simp
      done

    have "emeasure M (\<Union>i. if i \<in> to_nat_on X`X then ?C (from_nat_into X i) else {}) =
        (\<Sum>i. emeasure M (if i \<in> to_nat_on X`X then ?C (from_nat_into X i) else {}))"
      by (intro suminf_emeasure[symmetric]) (auto simp: disjoint cyl_def sets_eq S_seq_def)
    also have "\<dots> = (\<Sum>i. emeasure N (if i \<in> to_nat_on X`X then ?C (from_nat_into X i) else {}))"
      using F by (intro arg_cong[where f=suminf] ext) (force intro!: eq simp: less_Suc_eq_le)
    also have "\<dots> = emeasure N (\<Union>i. if i \<in> to_nat_on X`X then ?C (from_nat_into X i) else {})"
      by (intro suminf_emeasure) (auto simp: disjoint cyl_def sets_eq S_seq_def)
    finally show ?thesis
      unfolding * .
  next
    assume "\<not> (\<forall>i\<le>n. F i \<noteq> {})"
    then have *: "{\<omega> \<in> space S_seq. \<forall>i\<le>n. \<omega> i \<in> F i} = {}"
      by auto
    show ?thesis
      by (simp add: *)
  qed
qed

lemma (in wellorder) smallest:
  assumes "P i" obtains j where "P j" "\<And>i. i < j \<Longrightarrow> \<not> P i" "j \<le> i"
proof
  show "P (LEAST j. P j)" "\<And>i. i < (LEAST j. P j) \<Longrightarrow> \<not> P i" "(LEAST j. P j) \<le> i"
    using `P i` LeastI2_wellorder[of P i "\<lambda>j. j \<le> i", OF `P i`]
    by (auto intro: LeastI_ex dest: not_less_Least)
qed

lemma setsum_strict_mono_single: 
  fixes f :: "_ \<Rightarrow> 'a :: {comm_monoid_add,ordered_cancel_ab_semigroup_add}"
  shows "finite A \<Longrightarrow> a \<in> A \<Longrightarrow> f a < g a \<Longrightarrow> (\<And>a. a \<in> A \<Longrightarrow> f a \<le> g a) \<Longrightarrow> setsum f A < setsum g A"
  using setsum_strict_mono_ex1[of A f g] by auto

lemma zero_notin_Suc_image: "0 \<notin> Suc ` A" by auto

lemma Ex_nat_case_eq: "(\<exists>n. P n (nat_case s f n)) \<longleftrightarrow> P 0 s \<or> (\<exists>n. P (Suc n) (f n))"
  by (auto split: nat.split simp: gr0_conv_Suc)

lemma nat_case_in_funcset: "nat_case x f \<in> (UNIV \<rightarrow> X) \<longleftrightarrow> x \<in> X \<and> f \<in> UNIV \<rightarrow> X"
  by (auto simp: Pi_def split: nat.split)

lemma nat_case_idem: "nat_case (f 0) (\<lambda>x. f (Suc x)) = f"
  by (auto split: nat.split)

lemma all_plus_split: 
  fixes P :: "nat \<Rightarrow> bool"
  shows "(\<forall>i\<le>k + n. P i) \<longleftrightarrow> (\<forall>i\<le>n. P (i + k)) \<and> (\<forall>i<k. P i)"
  apply (intro iffI allI)
   apply force
  apply (case_tac "k \<le> i")
   apply (erule conjE)
   apply (erule_tac x="i - k" in allE)
   apply (auto simp add: not_less)
  done

lemma all_Suc_split: "(\<forall>i\<le>Suc n. P i) \<longleftrightarrow> (\<forall>i\<le>n. P (Suc i)) \<and> P 0"
  by (metis le_eq_less_or_eq less_Suc_eq_0_disj)

lemma all_less_Suc_split: "(\<forall>i<Suc n. P i) \<longleftrightarrow> (\<forall>i<n. P (Suc i)) \<and> P 0"
  by (metis less_Suc_eq_0_disj)

lemma nat_boundary_cases[case_names less add]:
  fixes P :: bool and n l :: nat
  shows "(n < l \<Longrightarrow> P) \<Longrightarrow> (\<And>k. n = k + l \<Longrightarrow> P) \<Longrightarrow> P"
  using less_imp_add_positive[of l n]
  by (cases n l rule: linorder_cases) auto

lemma
  fixes S :: "'s set"
  assumes S: "countable S"
  assumes P[measurable]: "\<And>s. s \<in> S \<Longrightarrow> Measurable.pred M (P s)"
  shows sets_Collect_All_S[measurable (raw)]: "Measurable.pred M (\<lambda>x. \<forall>s\<in>S. P s x)" (is ?ALL)
    and sets_Collect_Ex_S[measurable (raw)]: "Measurable.pred M (\<lambda>x. \<exists>s\<in>S. P s x)" (is ?EX)
proof -
  let ?f = "from_nat_into S"
  have *: 
    "\<And>x. (\<forall>s\<in>S. P s x) \<longleftrightarrow> (\<forall>n. ?f n \<in> S \<longrightarrow> P (if ?f n \<in> S then ?f n else undefined) x)"
    "\<And>x. (\<exists>s\<in>S. P s x) \<longleftrightarrow> (\<exists>n. ?f n \<in> S \<and> P (if ?f n \<in> S then ?f n else undefined) x)"
    using subset_range_from_nat_into[OF S] by auto
  from subset_range_from_nat_into[OF S] show ?ALL ?EX unfolding * by measurable
qed

lemma measurable_compose_countable:
  assumes S[simp]: "countable S"
  assumes g[measurable]: "g \<in> measurable M (count_space S)"
  assumes f[measurable]: "\<And>i. i \<in> S \<Longrightarrow> f i \<in> measurable M N"
  shows "(\<lambda>x. f (g x) x) \<in> measurable M N"
  unfolding measurable_def
proof safe
  fix x assume "x \<in> space M" then show "f (g x) x \<in> space N"
    using measurable_space[OF f] measurable_space[OF g] by auto
next
  fix A assume [measurable]: "A \<in> sets N"
  have *: "(\<lambda>x. f (g x) x) -` A \<inter> space M = {x\<in>space M. \<exists>s\<in>S. g x = s \<and> f s x \<in> A}"
    using measurable_space[OF g] by auto
  show "(\<lambda>x. f (g x) x) -` A \<inter> space M \<in> sets M"
    unfolding * by measurable
qed


lemma (in finite_measure) finite_measure_not_0_imp_not_0:
    "measure M X \<noteq> 0 \<Longrightarrow> Y \<in> sets M \<Longrightarrow> X \<subseteq> Y \<Longrightarrow> measure M Y \<noteq> 0"
  using finite_measure_mono[of X Y] measure_nonneg[of M Y] measure_nonneg[of M X]
  by auto

lemma emeasure_distr_Collect:
  "f \<in> measurable M N \<Longrightarrow> Measurable.pred N P \<Longrightarrow> emeasure (distr M N f) {x\<in>space N. P x} = emeasure M {x\<in>space M. P (f x)}"
  by (subst emeasure_distr) (auto intro!: arg_cong[where f="emeasure M"] dest: measurable_space)

lemma emeasure_distr_Collect':
  "f \<in> measurable M N \<Longrightarrow> Measurable.pred N P \<Longrightarrow> emeasure (distr M N f) {x\<in>space (distr M N f). P x} = emeasure M {x\<in>space M. P (f x)}"
  by (subst emeasure_distr) (auto intro!: arg_cong[where f="emeasure M"] dest: measurable_space)

lemma measure_not_0: "measure M X \<noteq> 0 \<longleftrightarrow> 0 < measure M X"
  using measure_nonneg[of M X] by auto

lemma someI_in: "S \<noteq> {} \<Longrightarrow> (SOME i. i \<in> S) \<in> S"
  by (rule someI_ex) auto

lemma finite_sequence:
  assumes "countable X" obtains F where "\<And>i. F i \<subseteq> X" "incseq F" "\<And>i. finite (F i)" "(\<Union>i. F i) = X"
proof -
  show thesis
    apply (rule that[of "\<lambda>i. if X = {} then {} else from_nat_into X ` {..i}"])
    apply (auto simp: incseq_def image_iff Ball_def intro: from_nat_into split: split_if_asm)
  proof -
    fix x n assume "x \<in> X" "\<forall>i m. m \<le> i \<longrightarrow> x \<noteq> from_nat_into X m"
    with from_nat_into_surj[OF `countable X` `x \<in> X`]
    show False
      by auto
  qed
qed


lemma sets_PIM_cong: assumes "I = J" "(\<And>i. i \<in> J \<Longrightarrow> sets (M i) = sets (N i))" shows "sets (PiM I M) = sets (PiM J N)"
  using sets_eq_imp_space_eq[OF assms(2)] assms
  apply (simp add: sets_PiM prod_algebra_def prod_emb_def cong: PiE_cong)
  apply (intro arg_cong2[where f=sigma_sets] refl image_cong arg_cong[where f=Collect] ext prod.case_cong conj_cong)
  apply (simp add: Pi_iff)
  apply (intro ball_cong refl arg_cong2[where f="op :"] assms)
  apply auto
  done

locale finite_pmf = fixes S :: "'a set" and p :: "'a \<Rightarrow> real"
  assumes finite_S: "finite S" and nonneg_p[simp]: "\<And>i. i \<in> S \<Longrightarrow> 0 \<le> p i" and distr_p: "(\<Sum>i\<in>S. p i) = 1"
begin

definition "M = point_measure S p"

lemma sets_M[simp]: "sets M = Pow S" and space_M[simp]:  "space M = S"
  by (simp_all add: M_def point_measure_def)

lemma emeasure_M[simp]: "A \<subseteq> S \<Longrightarrow> emeasure M A = (\<Sum>i\<in>A. p i)"
  unfolding M_def by (subst emeasure_point_measure_finite) (auto simp: finite_S nonneg_p)

lemma measure_M[simp]: "A \<subseteq> S \<Longrightarrow> measure M A = (\<Sum>i\<in>A. p i)"
  unfolding M_def measure_def by (subst emeasure_point_measure_finite) (auto simp: finite_S nonneg_p)

lemma positive_integral_M[simp]: "(\<And>i. i \<in> S \<Longrightarrow> 0 \<le> f i) \<Longrightarrow> integral\<^sup>P M f = (\<Sum>i\<in>S. f i * p i)"
  unfolding M_def by (subst positive_integral_point_measure_finite) (auto simp: finite_S nonneg_p ac_simps)

lemma lebesgue_integral_M[simp]: "integral\<^sup>L M f = (\<Sum>i\<in>S. f i * p i)"
  unfolding M_def by (subst lebesgue_integral_point_measure_finite) (auto simp: finite_S nonneg_p ac_simps)

lemma prob_space_M: "prob_space M"
  by rule (simp add: distr_p one_ereal_def)

sublocale prob_space M
  by (rule prob_space_M)

end

lemma Gcd_eq_one: "1 \<in> A \<Longrightarrow> Gcd A = (1::nat)"
  by (auto intro!: gcd_lcm_complete_lattice_nat.Inf_eqI)

lemma ereal_one_divide_invese: "0 \<le> x \<Longrightarrow> 1 / inverse x = (x::ereal)"
  by (cases x) (auto simp: divide_ereal_def)

end
