(*<*)
theory Overview
  imports Classifying_Markov_Chains
begin

context Discrete_Markov_Kernel
begin

(*>*)

section {* Introduction *}

text {*

The examples from Fig. 1 (a) and (b) in~\cite{hoelzl2013classifying} are formalized in 
Section~\ref{ex:A} and Section~\ref{ex:B}.

*}

subsection {* Constant table *}

text {*

In the paper~\cite{hoelzl2013classifying} some of the constants where renamed for presentation
purposes. 

\begin{center}
\begin{tabular}{ll}
\textbf{Paper} & \textbf{Isabelle/HOL theories} \\
$\mathbb{B}$ & @{typ bool} \\
$\mathbb{N}$ & @{typ nat} \\
$\mathbb{R}$ & @{typ real} \\
$\overline{\mathbb{R}}$ & @{typ ereal} \\
$\omega \circ_n \omega$ & @{term "comb_seq n \<omega> \<omega>"} \\
$s{\cdot}\omega$         & @{term "nat_case s \<omega>"} \\
$\chi_A\;x$             & @{term "indicator A x"} \\
$f'\;z \xrightarrow{z \to 1^-} y$ & @{term [eta_contract=false] "((\<lambda>z. f' z) ---> y) (at_left (1::real))"} \\
$X\;n \xrightarrow{n \to \infty} y$ & @{term [eta_contract=false] "(\<lambda>n. X n) ----> y"} \\
$\textit{count}$ & @{const count_space} \\
$\textit{point}$ & @{const point_measure} \\
$\Pi_{i \in I} M\;i$ & @{term [eta_contract=false] "\<Pi>\<^sub>M i \<in> I. M i"} \\
$\textit{trace}$ & @{term [names_short=true] path} \\
$\mathcal{P}_x$ & @{term "paths x"} \\
$(x, y) \in \textit{accessible}$ & @{term "accessible x y"} \\
$\textit{essential}$ & @{term essential_class} \\
$G_g$ & @{term "gf_G"} \\
$M_g$ & @{term "gf_U'"} \\
$U_g$ & @{term "gf_U"} \\
\end{tabular}
\end{center}

*}

subsection {* List of Definitions and Theorems *}

text {*

\begin{itemize}
  \item \texttt{locale}~Discrete-Time-Markov-Chain -- Section~\ref{sec:stochasticprocess}
  \item \texttt{locale}~Discrete-Markov-Kernel -- Section~\ref{sec:discretemarkovkernel}
  \item Definition of $\textit{trace}$ -- @{const path} in Section~\ref{sec:path}
  \item Definition~1 (Trace space) $\mathcal{P}_x$ -- @{const paths} in Section~\ref{sec:paths}
  \item Lemma~2 (Splitting rules) in Section~\ref{sec:splitting_rules}
  \item Lemma~3 (Chapman-Kolmogorov) in Section~\ref{sec:kologorov}
  \item Theorem~4 (The trace space $\mathcal{P}_x$ is a Markov chain) in Section~\ref{sec:kerneldtmc}
  \item Definition~5 (Markov kernel from the stochastic process) in Section~\ref{sec:mcstoch}
  \item Theorem~6 (The Markov kernel $K'$ has the distribution of $X$) in Section~\ref{sec:KhasX}
  \item Definition 7 (Enabled states) in Section~\ref{sec:enabled}
  \item Theorem 8 in Section~\ref{sec:aeenabled}
  \item Definition 9 (Accessible states) in Section~\ref{sec:accessible}
  \item Theorem 10. in Section~\ref{sec:accessibleiff} (In the paper @{const p} is unfolded)
  \item Definition 11 (Communicating states) in Section~\ref{sec:communicating}
  \item Definition 12 (Essential class). in Section~\ref{sec:essential}
  \item Definition 13 (Recurrent states) in Section~\ref{sec:recurrent}
  \item Definition 14 (Hitting time) in Section~\ref{sec:hitting}
  \item Definition 15 (Average hitting time) is does not have it's own constant
  \item Definition 16 (Positive recurrent states) in Section~\ref{sec:posrecurrent}
  \item Definition 17 (Unbound quantities) in Section~\ref{sec:GUF} and Section~{sec:recurrent}
  \item Definition 18 (Time-bounded reachability probabilities) in Section~\ref{sec:timebound}
  \item Lemma 19 (Unbound quantities as series of time-bounded probabilities) in Section~\ref{sec:geqp}, \ref{sec:uequ}, and \ref{sec:mequ}
  \item Definition 20 (Generating functions for G, M , U, and F ) in Section~\ref{sec:genfun} and \ref{sec:hitting}
  \item Lemma 21 (Unbound quantities as limits of generating functions) in Section~\ref{sec:genfun} and \ref{sec:gfM}
  \item Theorem 22 (Relating recurrent and G) in Section~\ref{sec:recurrentG} and \ref{sec:gfGgfU}
  \item Corollary 23 (recurrent is invariant on irreducible classes) in Section~\ref{sec:invariantreccurrent}
  \item Lemma 24 (Relation between recurrent and H and U) in Section~\ref{sec:recurrent} and \ref{sec:recurrentHU}
  \item Theorem 25 (Recurrent classes are essential) in Section~\ref{sec:essential}
  \item Corollary 26 (Finite essential classes are recurrent) in Section~\ref{sec:essential}
  \item Lemma 27 (Relation between M and U) in Section~\ref{sec:gfM}
  \item Corollary 28 (Positive recurrent is invariant on irreducible classes) in Section~\ref{sec:posrecurrent}
  \item Definition 29 (Stationary distribution) in Section~\ref{sec:stationary}
  \item Theorem 31 (Stationary distribution implies positive recurrence) in Section~\ref{sec:sdimppr}
  \item Definition 32 (Aperiodic classes) in Section~\ref{sec:aperiodic}
  \item Lemma 33 (Aperiodic class implies non-zero p) in Section~\ref{sec:aperiodicp} \\
    Here instead of the quantifiers $\exists N.~\forall n \ge N.~P\;n$ we use @{term "eventually P sequentially"}
  \item Theorem 34 (Stationary distribution is asymptotic distribution) in Section~\ref{sec:limit}
\end{itemize}

*}

(*<*)
end

end
(*>*)

